/*****************************************************************************
 *
 * pciscc.c	This is the device driver for the PCISCC-4 card or any other
 *		board based on the Siemens PEB-20534H (DSCC-4) communication
 *              controller. The PCISCC-4 is a four-channel medium-speed (up
 *		to 10 respectively 52 Mbps/channel) synchronous serial
 *		interface controller with HDLC protocol processor and
 *		busmaster-DMA facilities.
 *
 * Author:	(c) 2001/2002 Steffen Koehler <dh1dm@qsl.net>
 *
 *****************************************************************************
 */

#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include "flexdrv.h"
#include "pciscc.h"
#include "pci.h"
#include "mem32.h"

/* ------------------------------------------------------------------------- */

#define	CHIP_TIMEOUT	2		/* chip timeout (55ms intervals) */
#define	XTAL_FREQ	19660000	/* on board crystal freq. */
#define MIN_FLAGS	32		/* minimum delay/tail bits */

/* ------------------------------------------------------------------------- */

char helptext[] =
"\nUsage: PCISCC [options]\n"
"Oprions: /c=<num> select a PCISCC card if more than one is present\n"
"                  default card number is 0\n"
"         /t=<ms>  set TX delay scaling, default value is 10 ms\n";

char version[] = "0.4a";

/* ------------------------------------------------------------------------- */

/* Convert far pointer to physical address */
#define FP_TO_PHYS(x)	((long)(((long)FP_SEG(x) << 4) + FP_OFF(x)))

/* global variables */
struct chipctl_t chipctl;
struct devctl_t devctl[4];
struct buffer_t dma_buffers;

/* processed flexnet RX/TX frames */
L1FRAME far *tx_frame = NULL;
L1FRAME far *rx_frame = NULL;

/* BIOS timer */
unsigned long far *bios_tic = MK_FP(0x40,0x6c);

/* ------------------------------------------------------------------------- */

/*
 * get a frame buffer from pre-allocated driver frame memory pool and
 * mark this frame buffer as used
 */
L1FRAME far * alloc_frame(void)
{
	int i;
	for (i=0; i < CFG_L1_BUF; i++) {
		if (!chipctl.dma_buffers->frm_pool[i].valid) {
			chipctl.dma_buffers->frm_pool[i].valid = 1;
			return &chipctl.dma_buffers->frm_pool[i].l1_frame;
		}
	}
	return NULL;
}

/* ------------------------------------------------------------------------- */

/*
 * find a previosly allocated frame buffer in the driver frame memory pool
 * and mark this frame buffer as unused
 */
void free_frame(L1FRAME far * frame)
{
	int i;
	for (i=0; i < CFG_L1_BUF; i++) {
		if (&chipctl.dma_buffers->frm_pool[i].l1_frame == frame) {
			chipctl.dma_buffers->frm_pool[i].valid = 0;
			break;
		}
	}
}

/* ------------------------------------------------------------------------- */

/*
 * get the number of available (unused) frame buffers in the driver
 * pre-allocated frame memory pool
 */
int avail_frames(void)
{
	int i;
	int cnt = 0;
	for (i=0; i < CFG_L1_BUF; i++)
		if (!chipctl.dma_buffers->frm_pool[i].valid) cnt++;
	return cnt;
}

/* ------------------------------------------------------------------------- */

/* initialize channel queues */
void pciscc_init_queues(int chan)
{
	struct rx_desc_t *first_rdp = NULL;
	struct rx_desc_t *curr_rdp = NULL;
	struct rx_desc_t *last_rdp = NULL;
	struct tx_desc_t *first_tdp = NULL;
	struct tx_desc_t *curr_tdp = NULL;
	struct tx_desc_t *last_tdp = NULL;
	unsigned int i;

	/* initialize TX/RX IQs */
	devctl[chan].iq_rx = chipctl.dma_buffers->iq_rx[chan];
	devctl[chan].iq_tx = chipctl.dma_buffers->iq_tx[chan];
	memset(devctl[chan].iq_rx, 0, 4*CFG_IQLEN);
	memset(devctl[chan].iq_tx, 0, 4*CFG_IQLEN);
	devctl[chan].iq_rx_next = devctl[chan].iq_rx;
	devctl[chan].iq_tx_next = devctl[chan].iq_tx;
	/*
	 * Prepare circular RX and TX descriptor queues ("FIFO" rings)
	 * Attention:
	 * This beast gets _very_ angry if you try to hand it a
	 * descriptor with a data length of 0. In fact it crashes
	 * the system by asserting /SERR or something.
	 */
	devctl[chan].dq_rx = chipctl.dma_buffers->dq_rx[chan];
	first_rdp = devctl[chan].dq_rx;
	for (i=0, curr_rdp = first_rdp; i < CFG_RX_DESC; i++, curr_rdp++) {
		if (i > 0) {
			curr_rdp->prev = last_rdp;
			last_rdp->next = curr_rdp;
			last_rdp->nextptr = FP_TO_PHYS(curr_rdp);
		}
		disable(); curr_rdp->l1_frame = alloc_frame(); enable();
		curr_rdp->dataptr = FP_TO_PHYS(curr_rdp->l1_frame->frame);
		curr_rdp->flags = CFG_MTU*NO;
		curr_rdp->feptr = 0;
		curr_rdp->result = 0;
		last_rdp = curr_rdp;
	}
	last_rdp->next = first_rdp;			/* close ring structure */
	last_rdp->nextptr = FP_TO_PHYS(first_rdp);
	first_rdp->prev = last_rdp;
	devctl[chan].dq_rx_next = first_rdp;		/* first descriptor to be processed = "first" descriptor in chain */
	/* TX queue */
	devctl[chan].dq_tx = chipctl.dma_buffers->dq_tx[chan];
	first_tdp = devctl[chan].dq_tx;
	for (i=0, curr_tdp = first_tdp; i < CFG_TX_DESC; i++, curr_tdp++) {
		if (i > 0) {
			curr_tdp->prev = last_tdp;
			last_tdp->next = curr_tdp;
			last_tdp->nextptr = FP_TO_PHYS(curr_tdp);
		}
		curr_tdp->l1_frame = NULL;
		curr_tdp->dataptr = FP_TO_PHYS(chipctl.dma_buffers->dummy);
		curr_tdp->flags = FE | (8*NO);
		curr_tdp->result = 0;
		last_tdp = curr_tdp;
	}
	last_tdp->next = first_tdp;			/* close ring structure */
	last_tdp->nextptr = FP_TO_PHYS(first_tdp);
	first_tdp->prev = last_tdp;
	devctl[chan].dq_tx_last = first_tdp;		/* last descriptor to be transmitted */
	devctl[chan].dq_tx_cleanup = first_tdp;		/* first descriptor to be cleaned */
}

/* ------------------------------------------------------------------------- */

/* find best BRG N and M match for given bitrate */
static void pciscc_set_baud(int chan, long rate)
{
	long brg_best_n	= 0;
	long brg_best_m	= 0;
	long brg_tmp,m,n;
	unsigned long brg_error	= XTAL_FREQ;
	unsigned long brg_tmp_error;

	for (m=0; m<16; m++) {
		for (n=0; n<63; n++) {
			brg_tmp = (n+1)*(1<<m);
			brg_tmp_error = labs(XTAL_FREQ/brg_tmp - rate);
			if (brg_tmp_error < brg_error) {
				brg_best_m = m;
				brg_best_n = n;
				brg_error = brg_tmp_error;
			}
		}
	}
	WriteL(chipctl.io_base+SCCBASE[chan]+BRR,(brg_best_m*BRM) | (brg_best_n*BRN));
}

/* ------------------------------------------------------------------------- */

/* detect/initialize PCISCC chip, called from init_device() function */
int pciscc_chip_init(void)
{
	unsigned char pci_devfn;
	unsigned char pci_bus;
	unsigned int cmd;
	char hexstr[10];

	/* make sure PCI BIOS is avaliable */
	if(!PCIBiosPresent()) {
		pstr("PCISCC: No PCI bios present.\n");
		return 1;
	}
	pstr("PCISCC: Card "); pnum(chipctl.card_num); pstr(" selected.\n");
	if (!(FindPCIDevice(PCI_DEVICE_ID, PCI_VENDOR_ID, chipctl.card_num, &pci_bus, &pci_devfn))) {
		pstr("PCISCC: PEB-20534H (DSCC4) card not found.\n");
		return 1;
	}
	/* read chip configuration */
	ReadConfigurationDword(pci_bus, pci_devfn, PCI_CS_BASE_ADDRESS_0, &chipctl.io_base);
	ReadConfigurationByte(pci_bus, pci_devfn, PCI_CS_INTERRUPT_LINE, &chipctl.irq);
	/* enable chip fixups */
	ReadConfigurationWord(pci_bus, pci_devfn, PCI_CS_COMMAND, &cmd);
	cmd |= PCI_COMMAND_MASTER;
	cmd |= PCI_COMMAND_PARITY;
	cmd |= PCI_COMMAND_SERR;
	cmd &= ~PCI_COMMAND_FAST_BACK;
	WriteConfigurationWord(pci_bus, pci_devfn, PCI_CS_COMMAND, cmd);
	WriteConfigurationByte(pci_bus, pci_devfn, PCI_CS_MASTER_LATENCY, 0xF8);
	/* show probed hardware parameters */
	pstr("PCISCC: Found PEB-20534H chip at mem=0x");
	ltoa(chipctl.io_base, hexstr, 16); pstr(hexstr);
	pstr(", irq="); pnum(chipctl.irq); pstr(".\n");
	/* display TX delay scaling */
	pstr("PCISCC: TX delay scaling set to ");
	pnum(chipctl.txd_scale); pstr(" ms.\n");
	return 0;
}

/* ------------------------------------------------------------------------- */

/*
 * initialize chip, called from pciscc_channel_open() function
 * action sequency was carefully chosen, don't mess with it
 */
int pciscc_chip_open(void)
{
	unsigned int nvec, mask;
	unsigned long timer;
	unsigned long i;

	if (!chipctl.io_base || !chipctl.irq)
		return 0;
	if (chipctl.initialized)
		return 1;
	/* align buffer memory address */
	chipctl.dma_buffers = (void*)(((long)&dma_buffers & 0xfffffffc) + 4);
	/* clear frame buffer pool */
	for (i=0; i < CFG_L1_BUF; i++)
		chipctl.dma_buffers->frm_pool[i].valid = 0;
	/* init rx frame queue */
	chipctl.rx_get = chipctl.rx_put = chipctl.rx_queue;
	/* install IRQ service routine */
	if (chipctl.irq & 8)
		nvec = chipctl.irq + 0x68;
	else
		nvec = chipctl.irq + 0x08;
	chipctl.oldvect = getvect(nvec);
	setvect(nvec, pciscc_isr);
	/* save PIC state */
	chipctl.oldpic[0] = inportb(PIC0MASK);
	chipctl.oldpic[1] = inportb(PIC1MASK);
	/* enable IRQ line at interrupt controllers */
	mask = 1 << chipctl.irq;
	outportb(PIC0MASK, inportb(PIC0MASK) & ~mask);
	mask >>= 8;
	outportb(PIC1MASK, inportb(PIC1MASK) & ~mask);
	/* initialize peripheral queue */
	chipctl.iq_per = chipctl.dma_buffers->iq_per;
	memset(chipctl.iq_per, 0, 4*CFG_IQLEN);
	chipctl.iq_per_next = chipctl.iq_per;
	/* configuration interrupt queue */
	chipctl.iq_cfg = chipctl.dma_buffers->iq_cfg;
	memset(chipctl.iq_cfg, 0, 4*CFG_IQLEN);
	chipctl.iq_cfg_next = chipctl.iq_cfg;
	/* global hardware initialization */
	disable();
	WriteL(chipctl.io_base+GMODE, (4 * PERCFG) | (3 * LCD) | CMODE);
	WriteL(chipctl.io_base+IQPBAR,FP_TO_PHYS(chipctl.iq_per));
	WriteL(chipctl.io_base+IQCFGBAR,FP_TO_PHYS(chipctl.iq_cfg));
	WriteL(chipctl.io_base+IQLENR2,(((CFG_IQLEN/32)-1)*IQCFGLEN) | (((CFG_IQLEN/32)-1)*IQPLEN));
	WriteL(chipctl.io_base+FIFOCR1,((32/4)*TFSIZE0)
		| ((32/4)*TFSIZE1)
		| ((32/4)*TFSIZE2)
		| ((32/4)*TFSIZE3));
	WriteL(chipctl.io_base+FIFOCR2,((24/4)*TFRTHRES0)
		| ((24/4)*TFRTHRES1)
		| ((24/4)*TFRTHRES2)
		| ((24/4)*TFRTHRES3)
		| M4_0 | M4_1 | M4_2 | M4_3);
	WriteL(chipctl.io_base+FIFOCR3,CFG_FIFO_RX_T);
	WriteL(chipctl.io_base+FIFOCR4,(20*TFFTHRES0)
		| (20*TFFTHRES1)
		| (20*TFFTHRES2)
		| (20*TFFTHRES3));
	/* mask out all DMAC interrupts */
	WriteL(chipctl.io_base+CH0CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH1CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH2CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH3CFG,(MRFI | MTFI | MRERR | MTERR));
	/* all SCC cores in reset state */
	WriteL(chipctl.io_base+SCCBASE[0]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[1]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[2]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[3]+CCR0,0x00000000);
	/* mask out all SCC interrupts */
	WriteL(chipctl.io_base+SCCBASE[0]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[1]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[2]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[3]+IMR,0xffffffff);
	/* peripheral configuration */
	WriteL(chipctl.io_base+LCONF,(BTYP*3));
	WriteL(chipctl.io_base+SSCCON,0x00000000);
	WriteL(chipctl.io_base+SSCIM,0x00000000);
	WriteL(chipctl.io_base+GPDIR,0x000000ff);
	WriteL(chipctl.io_base+GPDATA,0x00000000);
	WriteL(chipctl.io_base+GPIM,0x00000000);
	enable();
	/* initialize configuration and peripheral IQs */
	chipctl.mailbox = MAILBOX_NONE;
	WriteL(chipctl.io_base+GCMDR, CFGIQCFG | CFGIQP | AR);
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (chipctl.mailbox) break;
	}
	if (chipctl.mailbox == MAILBOX_OK) {	/* mailbox was written by isr */
		chipctl.initialized = 1;
		return 1;
	}
	/* global configuration to reset state */
	disable();
	WriteL(chipctl.io_base+GMODE, (4 * PERCFG) | (3 * LCD) | CMODE | OSCPD);
	/* mask all DMAC interrupts */
	WriteL(chipctl.io_base+CH0CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH1CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH2CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH3CFG,(MRFI | MTFI | MRERR | MTERR));
	/* SCC cores to reset state */
	WriteL(chipctl.io_base+SCCBASE[0]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[1]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[2]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[3]+CCR0,0x00000000);
	/* mask all SCC interrupts */
	WriteL(chipctl.io_base+SCCBASE[0]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[1]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[2]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[3]+IMR,0xffffffff);
	enable();
	/* restore PIC state */
	outportb(PIC0MASK, chipctl.oldpic[0]);
	outportb(PIC1MASK, chipctl.oldpic[1]);
	/* restore previous vector */
	if (chipctl.irq & 8)
		nvec = chipctl.irq + 0x68;
	else
		nvec = chipctl.irq + 0x08;
	setvect(nvec, chipctl.oldvect);
	chipctl.iq_per_next = NULL;
	chipctl.iq_cfg_next = NULL;
	return 0;
}

/* ------------------------------------------------------------------------- */

/*
 * close chip, called when last device (channel) of a chip was closed.
 * don't mess up.
 */
void pciscc_chip_close(void)
{
	int nvec;

	if (!chipctl.initialized)
		return;
	chipctl.initialized = 0;
	/* global configuration to reset state */
	disable();
	WriteL(chipctl.io_base+GMODE, (4 * PERCFG) | (3 * LCD) | CMODE | OSCPD);
	/* mask all DMAC interrupts */
	WriteL(chipctl.io_base+CH0CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH1CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH2CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(chipctl.io_base+CH3CFG,(MRFI | MTFI | MRERR | MTERR));
	/* SCC cores to reset state */
	WriteL(chipctl.io_base+SCCBASE[0]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[1]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[2]+CCR0,0x00000000);
	WriteL(chipctl.io_base+SCCBASE[3]+CCR0,0x00000000);
	/* mask all SCC interrupts */
	WriteL(chipctl.io_base+SCCBASE[0]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[1]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[2]+IMR,0xffffffff);
	WriteL(chipctl.io_base+SCCBASE[3]+IMR,0xffffffff);
	enable();
	/* restore PIC state */
	outportb(PIC0MASK, chipctl.oldpic[0]);
	outportb(PIC1MASK, chipctl.oldpic[1]);
	/* restore previous vector */
	if (chipctl.irq & 8)
		nvec = chipctl.irq + 0x68;
	else
		nvec = chipctl.irq + 0x08;
	setvect(nvec, chipctl.oldvect);
	chipctl.iq_per_next = NULL;
	chipctl.iq_cfg_next = NULL;
}

/* ------------------------------------------------------------------------- */

/*
 * open one channel, the sequence of actions done here was carefully chosen,
 * don't mess with it unless you know exactly what you are doing...
 */
int pciscc_channel_open(int chan, long bitrate, int mode)
{
	unsigned long timer;
	unsigned long l;

	if (!bitrate)
		return 0;
	/* general chip init */
	if (!pciscc_chip_open()) {
		devctl[chan].stats.io_error++;
		return 0;
	}
	if ((bitrate == devctl[chan].bitrate) && (mode == devctl[chan].mode))
		return 1;
	/* reset channel configuration */
	pciscc_channel_close(chan);
	/* RX buffer space check */
	if (avail_frames() < CFG_RX_DESC) {
		devctl[chan].stats.rx_overrun++;
		return 0;
	}
	/* setup bitrate/mode for flexnet internal use */
	devctl[chan].mode = mode;
	devctl[chan].bitrate = bitrate;
	devctl[chan].scale = 614400 / bitrate;
	/* init interrupt and descriptor queues */
	pciscc_init_queues(chan);
	/* stop SCC */
	disable();
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR1,0);
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR2,0);
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR0,0);
	devctl[chan].ccr0 = devctl[chan].ccr1 = devctl[chan].ccr2 = 0;
	WriteL(chipctl.io_base+SCCBASE[chan]+TIMR,0);
	/* set IQ lengths and base addresses */
	l = ReadL(chipctl.io_base+IQLENR1);
	switch (chan) {
	case 0:	WriteL(chipctl.io_base+CH0CFG,MRFI | MTFI | MRERR | MTERR);
		WriteL(chipctl.io_base+IQSCC0RXBAR,FP_TO_PHYS(devctl[chan].iq_rx));
		WriteL(chipctl.io_base+IQSCC0TXBAR,FP_TO_PHYS(devctl[chan].iq_tx));
		l &= 0x0fff0fff;
		l |= (((CFG_IQLEN/32)-1)*IQSCC0RXLEN) | (((CFG_IQLEN/32)-1)*IQSCC0TXLEN);
		break;
	case 1:	WriteL(chipctl.io_base+CH1CFG,MRFI | MTFI | MRERR | MTERR);
		WriteL(chipctl.io_base+IQSCC1RXBAR,FP_TO_PHYS(devctl[chan].iq_rx));
		WriteL(chipctl.io_base+IQSCC1TXBAR,FP_TO_PHYS(devctl[chan].iq_tx));
		l &= 0xf0fff0ff;
		l |= (((CFG_IQLEN/32)-1)*IQSCC1RXLEN) | (((CFG_IQLEN/32)-1)*IQSCC1TXLEN);
		break;
	case 2:	WriteL(chipctl.io_base+CH2CFG,MRFI | MTFI | MRERR | MTERR);
		WriteL(chipctl.io_base+IQSCC2RXBAR,FP_TO_PHYS(devctl[chan].iq_rx));
		WriteL(chipctl.io_base+IQSCC2TXBAR,FP_TO_PHYS(devctl[chan].iq_tx));
		l &= 0xff0fff0f;
		l |= (((CFG_IQLEN/32)-1)*IQSCC2RXLEN) | (((CFG_IQLEN/32)-1)*IQSCC2TXLEN);
		break;
	case 3:	WriteL(chipctl.io_base+CH3CFG,MRFI | MTFI | MRERR | MTERR);
		WriteL(chipctl.io_base+IQSCC3RXBAR,FP_TO_PHYS(devctl[chan].iq_rx));
		WriteL(chipctl.io_base+IQSCC3TXBAR,FP_TO_PHYS(devctl[chan].iq_tx));
		l &= 0xfff0fff0;
		l |= (((CFG_IQLEN/32)-1)*IQSCC3RXLEN) | (((CFG_IQLEN/32)-1)*IQSCC3TXLEN);
		break;
	}
	WriteL(chipctl.io_base+IQLENR1,l);
	enable();
	chipctl.mailbox = MAILBOX_NONE;
	WriteL(chipctl.io_base+GCMDR, AR	/* Action Request */
		| (chan == 0 ? (CFGIQSCC0RX | CFGIQSCC0TX) : 0)
		| (chan == 1 ? (CFGIQSCC1RX | CFGIQSCC1TX) : 0)
		| (chan == 2 ? (CFGIQSCC2RX | CFGIQSCC2TX) : 0)
		| (chan == 3 ? (CFGIQSCC3RX | CFGIQSCC3TX) : 0));
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (chipctl.mailbox) break;
	}
	if (chipctl.mailbox != MAILBOX_OK) {	/* mailbox was written by isr */
		devctl[chan].stats.io_error++;
	}
	/* initialize channel's SCC core */
	disable();
	devctl[chan].ccr0 = PU | VIS | PSD;
	devctl[chan].ccr0 |= (mode & MODE_z) ? 0 : 2*SC;
	devctl[chan].ccr0 |= (mode & MODE_r) ? 0 : 6*CM;
	devctl[chan].ccr0 |= (mode & MODE_t) ? 0 : SSEL;
	devctl[chan].ccr1 = ODS | ICD | FCTS | MDS1 | SFLAG;
	devctl[chan].ccr2 = RAC | (3*RFTH) | ITF;
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR0,devctl[chan].ccr0);
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR1,devctl[chan].ccr1);
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR2,devctl[chan].ccr2);
	pciscc_set_baud(chan,(mode & MODE_r) ? bitrate : 16*bitrate);
	WriteL(chipctl.io_base+SCCBASE[chan]+RLCR,RCE | (CFG_MTU*RL));
	/*
	 * all sent | tx device underrun | timer int | tx message repeat |
	 * tx pool ready | rx device overflow | receive FIFO overflow |
	 * frame length exceeded => interrupt mask register
	 */
	WriteL(chipctl.io_base+SCCBASE[chan]+IMR,~(ALLS | XDU | TIN | XMR | XPR | RDO | RFO | FLEX));
	enable();
	/* wait until command executing (CEC) is clear */
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		l = ReadL(chipctl.io_base+SCCBASE[chan]+STAR);
		if (!(l & CEC)) break;
	}
	/* RX and TX SCC reset */
	WriteL(chipctl.io_base+SCCBASE[chan]+CMDR,RRES | XRES);
	devctl[chan].tx_mailbox = 0xffffffff;
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (devctl[chan].tx_mailbox != 0xffffffff) break;
	}
	if ((devctl[chan].tx_mailbox & 0x03ffffff) != 0x02001000) {	/* mailbox was written by isr */
		devctl[chan].stats.io_error++;
	}
	/* initialize DMAC channel's RX */
	disable();
	switch (chan) {
	case 0:	WriteL(chipctl.io_base+CH0CFG,IDR);
		WriteL(chipctl.io_base+CH0BRDA,FP_TO_PHYS(devctl[chan].dq_rx));
		WriteL(chipctl.io_base+CH0FRDA,FP_TO_PHYS(devctl[chan].dq_rx));
		WriteL(chipctl.io_base+CH0LRDA,FP_TO_PHYS(devctl[chan].dq_rx_next->prev->prev));
		break;
	case 1:	WriteL(chipctl.io_base+CH1CFG,IDR);
		WriteL(chipctl.io_base+CH1BRDA,FP_TO_PHYS(devctl[chan].dq_rx));
		WriteL(chipctl.io_base+CH1FRDA,FP_TO_PHYS(devctl[chan].dq_rx));
		WriteL(chipctl.io_base+CH1LRDA,FP_TO_PHYS(devctl[chan].dq_rx_next->prev->prev));
		break;
	case 2:	WriteL(chipctl.io_base+CH2CFG,IDR);
		WriteL(chipctl.io_base+CH2BRDA,FP_TO_PHYS(devctl[chan].dq_rx));
		WriteL(chipctl.io_base+CH2FRDA,FP_TO_PHYS(devctl[chan].dq_rx));
		WriteL(chipctl.io_base+CH2LRDA,FP_TO_PHYS(devctl[chan].dq_rx_next->prev->prev));
		break;
	case 3:	WriteL(chipctl.io_base+CH3CFG,IDR);
		WriteL(chipctl.io_base+CH3BRDA,FP_TO_PHYS(devctl[chan].dq_rx));
		WriteL(chipctl.io_base+CH3FRDA,FP_TO_PHYS(devctl[chan].dq_rx));
		WriteL(chipctl.io_base+CH3LRDA,FP_TO_PHYS(devctl[chan].dq_rx_next->prev->prev));
		break;
	}
	enable();
	chipctl.mailbox = MAILBOX_NONE;
	WriteL(chipctl.io_base+GCMDR,AR);
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (chipctl.mailbox) break;
	}
	if (chipctl.mailbox != MAILBOX_OK)	/* mailbox was written by isr */
		devctl[chan].stats.io_error++;
	/* mask all DMAC interrupts (needed) */
	switch (chan) {
	case 0:	WriteL(chipctl.io_base+CH0CFG,MRFI | MTFI | MRERR | MTERR);
		break;
	case 1:	WriteL(chipctl.io_base+CH1CFG,MRFI | MTFI | MRERR | MTERR);
		break;
	case 2:	WriteL(chipctl.io_base+CH2CFG,MRFI | MTFI | MRERR | MTERR);
		break;
	case 3:	WriteL(chipctl.io_base+CH3CFG,MRFI | MTFI | MRERR | MTERR);
		break;
	}
	/* wait until command executing (CEC) is clear */
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		l = ReadL(chipctl.io_base+SCCBASE[chan]+STAR);
		if (!(l & CEC)) break;
	}
	/* SCC core TX reset (again) */
	WriteL(chipctl.io_base+SCCBASE[chan]+CMDR,XRES);
	devctl[chan].tx_mailbox = 0xffffffff;
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (devctl[chan].tx_mailbox != 0xffffffff) break;
	}
	if ((devctl[chan].tx_mailbox & 0x03ffffff) != 0x02001000) {	/* mailbox was written by isr */
		devctl[chan].stats.io_error++;
	}
	/*
	 * initialize DMAC's TX channel, FI must stay masked all the time
	 * even during operation, see device errata 03/99
	 */
	switch (chan) {
	case 0:	WriteL(chipctl.io_base+CH0CFG,IDT | MTFI);
		WriteL(chipctl.io_base+CH0BTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		WriteL(chipctl.io_base+CH0FTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		WriteL(chipctl.io_base+CH0LTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		break;
	case 1:	WriteL(chipctl.io_base+CH1CFG,IDT | MTFI);
		WriteL(chipctl.io_base+CH1BTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		WriteL(chipctl.io_base+CH1FTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		WriteL(chipctl.io_base+CH1LTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		break;
	case 2:	WriteL(chipctl.io_base+CH2CFG,IDT | MTFI);
		WriteL(chipctl.io_base+CH2BTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		WriteL(chipctl.io_base+CH2FTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		WriteL(chipctl.io_base+CH2LTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		break;
	case 3:	WriteL(chipctl.io_base+CH3CFG,IDT | MTFI);
		WriteL(chipctl.io_base+CH3BTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		WriteL(chipctl.io_base+CH3FTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		WriteL(chipctl.io_base+CH3LTDA,FP_TO_PHYS(devctl[chan].dq_tx));
		break;
	}
	chipctl.mailbox = MAILBOX_NONE;
	WriteL(chipctl.io_base+GCMDR,AR);
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (chipctl.mailbox) break;
	}
	if (chipctl.mailbox == MAILBOX_OK)	/* mailbox was written by isr */
		devctl[chan].txstate=TX_IDLE;
	else
		devctl[chan].stats.io_error++;
	devctl[chan].tx_enqueued = 0;
	devctl[chan].initialized = 1;
	return 1;
}

/* ------------------------------------------------------------------------- */

/* close one channel - don't mess with it either */
void pciscc_channel_close(int chan)
{
	struct rx_desc_t *rdp;
	unsigned long timer;
	unsigned long l;

	if (!devctl[chan].initialized)
		return;
	devctl[chan].initialized=0;
	/* at first stop timer */
	WriteL(chipctl.io_base+SCCBASE[chan]+TIMR,0);
	/* wait until command executing (CEC) is clear */
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		l = ReadL(chipctl.io_base+SCCBASE[chan]+STAR);
		if (!(l & CEC)) break;
	}
	/* RX and TX SCC reset */
	WriteL(chipctl.io_base+SCCBASE[chan]+CMDR,RRES | XRES);
	devctl[chan].tx_mailbox = 0xffffffff;
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (devctl[chan].tx_mailbox != 0xffffffff) break;
	}
	if ((devctl[chan].tx_mailbox & 0x03ffffff) != 0x02001000) {	/* mailbox was written by isr */
		devctl[chan].stats.io_error++;
	}
	/* stop SCC core */
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR1,0);
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR2,0);
	WriteL(chipctl.io_base+SCCBASE[chan]+CCR0,0);
	devctl[chan].ccr0 = devctl[chan].ccr1 = devctl[chan].ccr2 = 0;
	devctl[chan].bitrate = devctl[chan].mode = 0;
	/*
	 * Give the isr some time to "refill" the rx-dq
	 * we _MUST_ guarantee that the DMAC-RX is _NOT_ in
	 * hold state when issuing the RESET command, otherwise the DMAC
	 * will crash. (DSCC-4 Rev. <= 2.1)
	 * In addition to that we may only issue a RESET if channel is
	 * currently really initialized, otherwise something horrible will
	 * result.
	 */
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT);
	/* OK, now we should be ready to put the DMAC into reset state */
	switch (chan) {
	case 0:	WriteL(chipctl.io_base+CH0CFG,RDR | RDT | MRFI | MTFI | MRERR | MTERR);
		break;
	case 1:	WriteL(chipctl.io_base+CH1CFG,RDR | RDT | MRFI | MTFI | MRERR | MTERR);
		break;
	case 2:	WriteL(chipctl.io_base+CH2CFG,RDR | RDT | MRFI | MTFI | MRERR | MTERR);
		break;
	case 3: WriteL(chipctl.io_base+CH3CFG,RDR | RDT | MRFI | MTFI | MRERR | MTERR);
		break;
	}
	chipctl.mailbox = MAILBOX_NONE;
	WriteL(chipctl.io_base+GCMDR,AR);
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (chipctl.mailbox) break;
	}
	if (chipctl.mailbox != MAILBOX_OK)	/* mailbox was written by isr */
		devctl[chan].stats.io_error++;
	devctl[chan].txstate = TX_RESET;
	/* clear IQs */
	l = ReadL(chipctl.io_base+IQLENR1);
	switch (chan) {
	case 0:	l &= 0x0fff0fff;
		WriteL(chipctl.io_base+IQLENR1,l);
		WriteL(chipctl.io_base+IQSCC0RXBAR,FP_TO_PHYS(chipctl.dma_buffers->dummy));
		WriteL(chipctl.io_base+IQSCC0TXBAR,FP_TO_PHYS(chipctl.dma_buffers->dummy));
		break;
	case 1: l &= 0xf0fff0ff;
		WriteL(chipctl.io_base+IQLENR1,l);
		WriteL(chipctl.io_base+IQSCC1RXBAR,FP_TO_PHYS(chipctl.dma_buffers->dummy));
		WriteL(chipctl.io_base+IQSCC1TXBAR,FP_TO_PHYS(chipctl.dma_buffers->dummy));
		break;
	case 2: l &= 0xff0fff0f;
		WriteL(chipctl.io_base+IQLENR1,l);
		WriteL(chipctl.io_base+IQSCC2RXBAR,FP_TO_PHYS(chipctl.dma_buffers->dummy));
		WriteL(chipctl.io_base+IQSCC2TXBAR,FP_TO_PHYS(chipctl.dma_buffers->dummy));
		break;
	case 3: l &= 0xfff0fff0;
		WriteL(chipctl.io_base+IQLENR1,l);
		WriteL(chipctl.io_base+IQSCC3RXBAR,FP_TO_PHYS(chipctl.dma_buffers->dummy));
		WriteL(chipctl.io_base+IQSCC3TXBAR,FP_TO_PHYS(chipctl.dma_buffers->dummy));
		break;
	}
	chipctl.mailbox = MAILBOX_NONE;
	WriteL(chipctl.io_base+GCMDR,AR
		| (chan == 0 ? (CFGIQSCC0RX | CFGIQSCC0TX) : 0)
		| (chan == 1 ? (CFGIQSCC1RX | CFGIQSCC1TX) : 0)
		| (chan == 2 ? (CFGIQSCC2RX | CFGIQSCC2TX) : 0)
		| (chan == 3 ? (CFGIQSCC3RX | CFGIQSCC3TX) : 0));
	timer = *bios_tic;
	while((*bios_tic - timer) < CHIP_TIMEOUT) {
		if (chipctl.mailbox) break;
	}
	if (chipctl.mailbox != MAILBOX_OK) {	/* mailbox was written by isr */
		devctl[chan].stats.io_error++;
	}
	rdp = devctl[chan].dq_rx;
	while(rdp->l1_frame) {
		free_frame(rdp->l1_frame);
		rdp->l1_frame = NULL;
		rdp = rdp->next;
	}
	devctl[chan].iq_rx_next=NULL;
	devctl[chan].iq_tx_next=NULL;
}

/* ------------------------------------------------------------------------- */

/* interrupt handler root */
void interrupt pciscc_isr(void)
{
	unsigned long status;
	unsigned long *iqp;
	int chan;

	disable();
	status = ReadL(chipctl.io_base+GSTAR);
	WriteL(chipctl.io_base+GSTAR,status);		/* ack' irq */
	/* process peripheral queue */
	iqp = chipctl.iq_per_next;
	while(iqp && *iqp) {
		*(iqp++) = 0;
		if (iqp == chipctl.iq_per+CFG_IQLEN)
			iqp = chipctl.iq_per;
	}
	chipctl.iq_per_next = iqp;
	/* process configuration queue */
	iqp = chipctl.iq_cfg_next;
	while(iqp && *iqp) {
		if ((*iqp) & ARACK)
			chipctl.mailbox = MAILBOX_OK;
		if ((*iqp) & ARF)
			chipctl.mailbox = MAILBOX_FAILURE;
		*(iqp++) = 0;
		if (iqp == chipctl.iq_cfg+CFG_IQLEN)
			iqp = chipctl.iq_cfg;
	}
	chipctl.iq_cfg_next = iqp;
	/* poll channel queues */
	for (chan=0; chan < 4; chan++) {
		/* process TX queue */
		pciscc_isr_txcleanup(chan);
		iqp = devctl[chan].iq_tx_next;
		while(iqp && *iqp) {
			if (*iqp & TIN) {
				/* timer interrupt */
				WriteL(chipctl.io_base+SCCBASE[chan]+TIMR,0);
				/* is there something in the TX queue ? */
				if (devctl[chan].tx_enqueued) {
					devctl[chan].txstate=TX_XMIT;
					switch (chan) {
					case 0:	WriteL(chipctl.io_base+CH0LTDA,FP_TO_PHYS(devctl[chan].dq_tx_last));
						break;
					case 1:	WriteL(chipctl.io_base+CH1LTDA,FP_TO_PHYS(devctl[chan].dq_tx_last));
						break;
					case 2:	WriteL(chipctl.io_base+CH2LTDA,FP_TO_PHYS(devctl[chan].dq_tx_last));
						break;
					case 3:	WriteL(chipctl.io_base+CH3LTDA,FP_TO_PHYS(devctl[chan].dq_tx_last));
						break;
					}
				} else {
					devctl[chan].txstate=TX_IDLE;
					devctl[chan].ccr1 &= ~RTS;
					WriteL(chipctl.io_base+SCCBASE[chan]+CCR1,devctl[chan].ccr1);
				}
			}
			if (*iqp & ALLS) {
				if (devctl[chan].tx_enqueued)
					devctl[chan].tx_enqueued--;
				if (!devctl[chan].tx_enqueued) {
					devctl[chan].txstate=TX_TAIL;
					WriteL(chipctl.io_base+SCCBASE[chan]+TIMR,MIN_FLAGS*TVALUE);
					WriteL(chipctl.io_base+SCCBASE[chan]+CMDR,STI);
				}
			}
			if (*iqp & XDU) {
				/* TX data underrun */
				devctl[chan].stats.tx_error++;
				devctl[chan].txstate=TX_STALL;
				devctl[chan].ccr1 &= ~RTS;
				WriteL(chipctl.io_base+SCCBASE[chan]+CCR1,devctl[chan].ccr1);
			}
			devctl[chan].tx_mailbox = *iqp;
			*(iqp++) = 0;
			if (iqp == devctl[chan].iq_tx+CFG_IQLEN)
				iqp = devctl[chan].iq_tx;
		}
		devctl[chan].iq_tx_next = iqp;
		/* process RX queue */
		pciscc_isr_receiver(chan);
		iqp = devctl[chan].iq_rx_next;
		while(iqp && *iqp) {
			/* statistics only */
			if ((*iqp & SCCIV_SCC) && (*iqp & SCCIV_RDO))
				devctl[chan].stats.rx_bufferoverflow++;
			if ((*iqp & SCCIV_SCC) && (*iqp & SCCIV_RFO))
				devctl[chan].stats.rx_bufferoverflow++;
			if ((*iqp & SCCIV_SCC) && (*iqp & SCCIV_FLEX))
				devctl[chan].stats.rx_bufferoverflow++;
			devctl[chan].rx_mailbox = *iqp;
			*(iqp++) = 0;
			if (iqp == devctl[chan].iq_rx+CFG_IQLEN)
				iqp = devctl[chan].iq_rx;
		}
		devctl[chan].iq_rx_next = iqp;
	}
	enable();
	if (status) {
		if (chipctl.irq & 8)			/* If irq > 7, send EOI to second PIC. */
			outp(PIC1MODE, PICEOI);
		outp(PIC0MODE, PICEOI);			/* Send EOI to first PIC. */
	} else
		chipctl.oldvect();
}

/* ------------------------------------------------------------------------- */

/* called by interrupt handler root when RX interrupt occurred */
void pciscc_isr_receiver(int chan)
{
	struct rx_desc_t *rdp;
	unsigned long status;
	unsigned char rdsb;
	unsigned int bno;
	unsigned int valid;
	unsigned int rx_cnt = 0;
	L1FRAME far * rx_new;

	if (!devctl[chan].initialized)
		return;
	rdp = devctl[chan].dq_rx_next;
	while (rdp->result & C) {
		status = rdp->result;
		bno = (status >> 16) & 0x1fff;
		valid = 1;				/* we assume frame valid */
		if ((status & RA) || (bno < 5) || (bno > CFG_MTU) || !(status & FE) || (rdp->feptr != FP_TO_PHYS(rdp))) {
			valid = 0;			/* aborted or invalid length */
		} else {
			rdsb = rdp->l1_frame->frame[bno-1];
			if (!(rdsb & SB_VFR))		/* incorrect bit length */
				valid = 0;
			if (rdsb & SB_RDO)		/* data overflow */
				valid = 0;
			if (!(rdsb & SB_CRC))		/* CRC error */
				valid = 0;
			if (rdsb & SB_RAB)		/* receive message aborted */
				valid = 0;
		}
		/* OK, this is a little bit tricky. We have to make sure
		 * that every descriptor has a buffer assigned. Thus we
		 * can only release a buffer to the link layer if we get
		 * a new one in turn before. */
		if (valid) {
			if ((rx_new = alloc_frame()) != NULL) {
				rdp->l1_frame->len = bno-1;
				rdp->l1_frame->kanal = chan;
				rdp->l1_frame->txdelay = 0;
				*(chipctl.rx_put++) = rdp->l1_frame;
				if (chipctl.rx_put == chipctl.rx_queue+CFG_L1_BUF)
					chipctl.rx_put = chipctl.rx_queue;
				rdp->l1_frame = rx_new;
				rdp->dataptr = FP_TO_PHYS(rx_new->frame);
				devctl[chan].stats.rx_frames++;
			} else
				devctl[chan].stats.rx_overrun++;
		}
		rdp->flags = CFG_MTU*NO;
		rdp->result = 0;
		rdp->feptr = 0;
		rdp = rdp->next;
		rx_cnt++;
	}
	devctl[chan].dq_rx_next = rdp;
	/*
	 * tell DMAC last available descriptor - keep up one
	 * descriptor space for security (paranoia) (...->prev->prev)
	 */
	if (rx_cnt) {
		switch (chan) {
		case 0:	WriteL(chipctl.io_base+CH0LRDA,FP_TO_PHYS(rdp->prev->prev));
			break;
		case 1:	WriteL(chipctl.io_base+CH1LRDA,FP_TO_PHYS(rdp->prev->prev));
			break;
		case 2:	WriteL(chipctl.io_base+CH2LRDA,FP_TO_PHYS(rdp->prev->prev));
			break;
		case 3:	WriteL(chipctl.io_base+CH3LRDA,FP_TO_PHYS(rdp->prev->prev));
			break;
		}
	}
}

/* ------------------------------------------------------------------------- */

/* called by IRQ handler root when a TX descriptor was completed */
void pciscc_isr_txcleanup(int chan)
{
	struct tx_desc_t *tdp;

	if (!devctl[chan].initialized)
		return;
	tdp = devctl[chan].dq_tx_cleanup;
	while (tdp->result & C) {
		/* clean up all (C)omplete descriptors */
		free_frame(tdp->l1_frame);
		tdp->l1_frame = NULL;
		tdp->flags = FE | (8*NO);	/* dummy */
		tdp->dataptr = FP_TO_PHYS(chipctl.dma_buffers->dummy);
		tdp->result = 0;
		tdp = tdp->next;
		devctl[chan].stats.tx_frames++;
	}
	devctl[chan].dq_tx_cleanup = tdp;
}

/* ------------------------------------------------------------------------- */

#pragma argsused
u16 init_device(i16 argc, char *argv[])
{
	char *ch;
	int i;

	pstr("\nBayCom/Flexnet PCISCC>4 driver V"); pstr(version);
        pstr(", (C)2001/2002 St.Koehler DH1DM\n");

	chipctl.card_num = 0; chipctl.txd_scale = 10;
	for (i = 1; i < argc; i++) {
		ch = argv[i];
		if (*ch == '-' || *ch == '/') ch++;
		if (!strncmpi(ch,"c=",2))
			chipctl.card_num = atoi(ch+2);
		else
		if (!strncmpi(ch,"t=",2))
			chipctl.txd_scale = atoi(ch+2);
		else
		if (*ch == '?' || *ch == 'h' || *ch == 'H') {
			pstr(helptext);
			return 99;
		} else {
			pstr("\nUnknown Option: '"); pstr(ch); pstr("\'\n");
			pstr(helptext);
			return 1;
		}
	}
	return pciscc_chip_init();
}

/* ------------------------------------------------------------------------- */

byte l1_ch_active(byte kanal)
{
	return 1;
}

/* ------------------------------------------------------------------------- */

void l1_exit(void)
{
	int chan;
	for (chan=0; chan < 4; chan++)
		pciscc_channel_close(chan);
	pciscc_chip_close();
}

/* ------------------------------------------------------------------------- */

u16 l1_get_ch_cnt(void)
{
	return 4;
}

/* ------------------------------------------------------------------------- */

byte l1_init_kanal(byte kanal, u16 baud, u16 mode)
{
	return pciscc_channel_open(kanal, (unsigned long)baud * 100, mode);
}

/* ------------------------------------------------------------------------- */

char far *l1_ident(byte kanal)
{
	return "PCISCC";
}

/* ------------------------------------------------------------------------- */

char far *l1_version(byte kanal)
{
	return version;
}

/* ------------------------------------------------------------------------- */

L1_STATISTICS far *l1_stat(byte kanal, byte delete)
{
	L1_STATISTICS *p = &devctl[kanal].stats;
	if (delete)
		p->tx_error = p->rx_overrun = p->rx_bufferoverflow =
		p->tx_frames = p->rx_frames = p->io_error = 0;
	return p;
}

/* ------------------------------------------------------------------------- */

/* report channel state (PTT/DCD) */
byte l1_ch_state(byte kanal)
{
/*
	 0x40   RxB RxBuffer ist nicht leer
	 0x20   PTT Sender ist an
	 0x10   DCD Empfnger ist aktiv
	 0x08   FDX Kanal ist Vollduplex, kann also immer empfangen
	 0x04   TBY Sender ist nicht bereit, z.B. wegen Calibrate
*/
	unsigned long l;
	unsigned int state = 0;

	if (!devctl[kanal].initialized)
		return 0;
	ReadL(chipctl.io_base+SCCBASE[kanal]+STAR);
	l = ReadL(chipctl.io_base+SCCBASE[kanal]+STAR);
	if (!(l & CD))
		state |= CH_DCD;
	if (chipctl.rx_get != chipctl.rx_put)
		state |= CH_RXB;
	switch(devctl[kanal].txstate) {
	case TX_DELAY:
	case TX_XMIT:
	case TX_TAIL:
		state |= CH_PTT;
		break;
	case TX_CAL:
		state |= CH_TBY | CH_PTT;
		break;
	case TX_RESET:
		state |= CH_TBY;
		break;
	}
	return state;
}

/* ------------------------------------------------------------------------- */

u16 l1_scale(byte kanal)
{
	return devctl[kanal].scale;
}

/* ------------------------------------------------------------------------- */

void l1_tx_calib(byte kanal, byte minutes)
{
	unsigned long timer;
	unsigned long bits;

	bits = devctl[kanal].bitrate * minutes * 60;
	if (devctl[kanal].txstate == TX_RESET) {
		devctl[kanal].stats.tx_error++;
		return;
	}
	if (devctl[kanal].txstate == TX_STALL) {
		WriteL(chipctl.io_base+SCCBASE[kanal]+CMDR,XRES);
		timer = *bios_tic;
		while((*bios_tic - timer) < CHIP_TIMEOUT);
		devctl[kanal].txstate = TX_IDLE;
	}
	if ((devctl[kanal].txstate == TX_CAL) && !minutes) {
		WriteL(chipctl.io_base+SCCBASE[kanal]+TIMR,MIN_FLAGS*TVALUE);
		WriteL(chipctl.io_base+SCCBASE[kanal]+CMDR,STI);
	}
	if (minutes) {
		devctl[kanal].ccr1 |= RTS;
		WriteL(chipctl.io_base+SCCBASE[kanal]+CCR1,devctl[kanal].ccr1);
		WriteL(chipctl.io_base+SCCBASE[kanal]+TIMR,bits*TVALUE);
		WriteL(chipctl.io_base+SCCBASE[kanal]+CMDR,STI);
		devctl[kanal].txstate = TX_CAL;
	}
}

/* ------------------------------------------------------------------------- */

void set_led(byte kanal, byte ledcode)
{
}

/* ------------------------------------------------------------------------- */

L1FRAME far *l1_rx_frame()
{
	disable();
	free_frame(rx_frame);
	rx_frame = NULL;
	if (chipctl.rx_get != chipctl.rx_put) {
		rx_frame = *(chipctl.rx_get++);
		if (chipctl.rx_get == chipctl.rx_queue+CFG_L1_BUF)
			chipctl.rx_get = chipctl.rx_queue;
	}
	enable();
	return rx_frame;
}

/* ------------------------------------------------------------------------- */

L1FRAME far *l1_get_framebuf(byte kanal)
{
	struct tx_desc_t *tdp;

	disable();
	tdp = devctl[kanal].dq_tx_last;
	if (devctl[kanal].initialized && !tx_frame)
		if ((tdp->next->l1_frame == NULL) && (tdp->next->next->l1_frame == NULL))
			tx_frame = alloc_frame();
	enable();
	return tx_frame;
}

/* ------------------------------------------------------------------------- */

/* enqueue/xmit pre-allocated TX frame buffer */
byte l1_tx_frame(void)
{
	struct tx_desc_t *tdp;
	unsigned long txd_bits;
	unsigned long timer;
	int chan;
	int enqueued = 0;

	/* TX frame assigned ? */
	if (!tx_frame)
		return 0;
	/* get channel number */
	chan = tx_frame->kanal;
	/* check if TX is in RESET state */
	if (devctl[chan].txstate == TX_RESET)
		devctl[chan].stats.tx_error++;
	/* if data underrun, reset TX now */
	if (devctl[chan].txstate == TX_STALL) {
		WriteL(chipctl.io_base+SCCBASE[chan]+CMDR,XRES);
		timer = *bios_tic;
		while((*bios_tic - timer) < CHIP_TIMEOUT);
		devctl[chan].txstate = TX_IDLE;
	}
	disable();
	/* check for TX queue space */
	tdp = devctl[chan].dq_tx_last->next;
	if ((tdp->l1_frame == NULL) && (tdp->next->l1_frame == NULL)) {
		/* insert TX frame buffer into descriptor queue */
		if ((devctl[chan].txstate == TX_IDLE) ||
		    (devctl[chan].txstate == TX_DELAY) ||
		    (devctl[chan].txstate == TX_XMIT) ||
		    (devctl[chan].txstate == TX_TAIL)) {
			tdp->l1_frame = tx_frame;
			tdp->dataptr = FP_TO_PHYS(tx_frame->frame);
			tdp->flags = FE | (NO * (tx_frame->len & 0x1fff));
			devctl[chan].dq_tx_last = tdp;
			devctl[chan].tx_enqueued++;
			enqueued++;
		}
		/* calculate tx keyup delay */
		txd_bits = devctl[chan].bitrate * tx_frame->txdelay * chipctl.txd_scale / 1000;
		if (txd_bits < MIN_FLAGS) txd_bits = MIN_FLAGS;
		/* TX was idle, key up and start txdelay */
		if (devctl[chan].txstate == TX_IDLE) {
			devctl[chan].txstate=TX_DELAY;
			devctl[chan].ccr1 |= RTS;
			WriteL(chipctl.io_base+SCCBASE[chan]+CCR1,devctl[chan].ccr1);
			WriteL(chipctl.io_base+SCCBASE[chan]+TIMR,txd_bits*TVALUE);
			WriteL(chipctl.io_base+SCCBASE[chan]+CMDR,STI);
		}
		/* update TX DMA register */
		if (devctl[chan].txstate == TX_XMIT) {
			switch (chan) {
			case 0:	WriteL(chipctl.io_base+CH0LTDA,FP_TO_PHYS(tdp));
				break;
			case 1:	WriteL(chipctl.io_base+CH1LTDA,FP_TO_PHYS(tdp));
				break;
			case 2:	WriteL(chipctl.io_base+CH2LTDA,FP_TO_PHYS(tdp));
				break;
			case 3:	WriteL(chipctl.io_base+CH3LTDA,FP_TO_PHYS(tdp));
				break;
			}
		}
	}
	if (!enqueued)
		free_frame(tx_frame);
	tx_frame = NULL;
	enable();
	return enqueued;
}

/*------------------------------------------------------------------------- */
